%{

#define LENGTH 36
#define DEBUG 0
#define MAX_STATE_STACK_SIZE 32
#define MAX_NUMBER_OF_STATES 16
#define VERSION "1.0 (1995-05-04)"

#ifdef THINK_C
#include <console.h>
#endif
#include <string.h>

enum letter_form {isolated, initial, medial, final, urdu_h};
enum letter_type {two_form, four_form, nothingT};
typedef struct {
	enum letter_form form_;
	enum letter_type type_;
	char kesh_;
    char alone_;
	char vowel[36]; 
	char cons[36]; } buffer_type;
buffer_type buffer[LENGTH];
int prev = 0, pos = 1, keshideh = 1, counter = 0, urdu = 0, debug = 0, display = 0;
const enum letter_form x0330[] = {isolated, final, final, isolated}, 
	x1221[] = {initial, medial, medial, initial},
	x0033[] = {isolated, isolated, final, final};
unsigned char form_string[5][5][2];
unsigned char arabic_delimiter = '';
unsigned char latin_delimiter = '';
unsigned char keshideh_string[10];
int state_stack_level = 0;
int state_stack[MAX_STATE_STACK_SIZE];

void purge(int);
void initialize_buffers(void);
void leave_arabic_mode(void);
void perform_CA(void);
void perform_shaddas(void);
void perform_lam_alifs(void);
void urduize(void);
void add_keshidehs(void);
void enter_arabic_mode(int);
void print_buffer(void);
void push_state(int);
void pop_state(int);
int yywrap(void);

%}

TWOFORM		("A"|"'a"|"'i"|"'A"|"d"|"dh"|"'d"|"r"|"z"|"zh"|"'r"|"'n"|"\"t"|"\"h"|"e"|"U"|"'u"|"E"|"x")

FOURFORM	("'y"|"b"|"t"|"th"|"p"|"'t"|"j"|"H"|"kh"|"ch"|"s"|"sh"|"S"|"D"|"T"|"Z"|"`"|"gh"|"f"|"q"|"v"|"k"|"g"|"l"|"m"|"n"|"h"|"I")

TWOFORM_ALONE		("-"("A"|"'a"|"'i"|"'A"|"d"|"dh"|"'d"|"r"|"z"|"zh"|"'r"|"'n"|"\"t"|"\"h"|"e"|"U"|"'u"|"E"|"x"))

FOURFORM_ALONE	("-"("'y"|"b"|"t"|"th"|"p"|"'t"|"j"|"H"|"kh"|"ch"|"s"|"sh"|"S"|"D"|"T"|"Z"|"`"|"gh"|"f"|"q"|"v"|"k"|"g"|"l"|"m"|"n"|"h"|"I"))

%x ARABIC
%x IGNORE
%x LATIN

%%

<*>^"#DEFINE ARABIC"(" ")*[^ \n]		{ arabic_delimiter = yytext[yyleng-1]; }

<*>^"#DEFINE LATIN"(" ")*[^ \n]		{ latin_delimiter = yytext[yyleng-1]; }

<*>^"%".*			{ ECHO; }

<*>^"#DEFINE "([^A]|(A[^R])|(AR[^A])|(ARA[^B])|(ARAB[^I])|(ARABI[^C])).*	;

<*>^"#URDU_MODE"	urdu=1;

<*>^"#ARABIC_MODE"	urdu=0;

<*>^"#KESHIDEH_MODE"	keshideh=1;

<*>^"#NO_KESHIDEH_MODE"	keshideh=0;

[\x80-\xFF][\x80-\xFF]	{ if (yytext[0]==yytext[1])	{
							if (arabic_delimiter == (unsigned char) yytext[0])  { enter_arabic_mode(1); }
									else ECHO; }
							else { 
								if (arabic_delimiter == (unsigned char) yytext[0])  { 
									enter_arabic_mode(0); unput(yytext[1]); }
								else {
									printf("%c",(unsigned char) yytext[0]);
									if (arabic_delimiter == (unsigned char) yytext[1])  { 
										enter_arabic_mode(0); }
									else printf("%c",(unsigned char) yytext[1]); 
									}
								}
							}

[\x80-\xFF]			{ if (arabic_delimiter == (unsigned char) yytext[0])  { enter_arabic_mode(0); }
					else ECHO; }
					
.|\n	ECHO;

<ARABIC>[\x80-\xFF][\x80-\xFF]	{ if (yytext[0]==yytext[1])	{
							if (arabic_delimiter == (unsigned char) yytext[0]) { leave_arabic_mode(); } 
									else ECHO; }
							else { 
								if (arabic_delimiter == (unsigned char) yytext[0]) { 
									leave_arabic_mode(); unput(yytext[1]); } 
								else if (latin_delimiter == (unsigned char) yytext[0]) { 
									purge(0); push_state(LATIN); unput(yytext[1]); } 
								else {
									printf("%c",(unsigned char) yytext[0]);
									if (arabic_delimiter == (unsigned char) yytext[1]) { 
										leave_arabic_mode();  } 
									else if (latin_delimiter == (unsigned char) yytext[1]) { 
										purge(0); push_state(LATIN); } 
									else printf("%c",(unsigned char) yytext[1]); 
									}
								}
							}

<ARABIC>[\x80-\xFF]			{ if (arabic_delimiter == (unsigned char) yytext[0]) { leave_arabic_mode(); } 
					else if (latin_delimiter == (unsigned char) yytext[0]) { purge(0); push_state(LATIN); } 
					else ECHO; }
					
<LATIN>[\x80-\xFF][\x80-\xFF]	{ if (yytext[0]==yytext[1])	{
							if (arabic_delimiter == (unsigned char) yytext[0])  { enter_arabic_mode(1); }
								else if (latin_delimiter == (unsigned char) yytext[0]) { pop_state(LATIN); } 
									else ECHO; }
							else { 
								if (arabic_delimiter == (unsigned char) yytext[0])  { 
									enter_arabic_mode(0); unput(yytext[1]); }
								else if (latin_delimiter == (unsigned char) yytext[0]) { pop_state(LATIN); unput(yytext[1]); } 
								else {
									printf("%c",(unsigned char) yytext[0]);
									if (arabic_delimiter == (unsigned char) yytext[1])  { 
										enter_arabic_mode(0); }
									else if (latin_delimiter == (unsigned char) yytext[1]) { pop_state(LATIN); } 
									else printf("%c",(unsigned char) (unsigned char) yytext[1]); 
									}
								}
							}

<LATIN>[\x80-\xFF]			{ if (arabic_delimiter == (unsigned char) yytext[0])  { enter_arabic_mode(0); }
					else if (latin_delimiter == (unsigned char) yytext[0]) { pop_state(LATIN); } 
					else ECHO; }
					
<LATIN>.|\n	ECHO;

<ARABIC>{TWOFORM}	{ buffer[counter].type_ = two_form; buffer[counter].alone_ = 0;
				buffer[counter].form_ = final;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,yytext); counter++; }

<ARABIC>{TWOFORM_ALONE}	{ buffer[counter].type_ = two_form; buffer[counter].alone_ = 1; 
				buffer[counter].form_ = final;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,yytext+1); counter++; }

<ARABIC>"\"A"	{ buffer[counter].type_ = two_form; buffer[counter].alone_ = 0; 
				buffer[counter].form_ = final;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,"AA"); counter++; }

<ARABIC>"-\"A"	{ buffer[counter].type_ = two_form; buffer[counter].alone_ = 1; 
				buffer[counter].form_ = final;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,"AA"); counter++; }

<ARABIC>"w"	{ buffer[counter].type_ = two_form; buffer[counter].alone_ = 0; 
				buffer[counter].form_ = final;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,"U"); counter++; }

<ARABIC>"-w"	{ buffer[counter].type_ = two_form; buffer[counter].alone_ = 1; 
				buffer[counter].form_ = final;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,"U"); counter++; }

<ARABIC>{FOURFORM} 	{	 buffer[counter].type_ = four_form; buffer[counter].alone_ = 0; 
				buffer[counter].form_ = medial;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,yytext); counter++; }

<ARABIC>{FOURFORM_ALONE} 	{	 buffer[counter].type_ = four_form; buffer[counter].alone_ = 1; 
				buffer[counter].form_ = medial;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,yytext+1); counter++; }

<ARABIC>"y" 	{	 buffer[counter].type_ = four_form; buffer[counter].alone_ = 0; 
				buffer[counter].form_ = medial;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,"I/"); counter++; }

<ARABIC>"-y" 	{	 buffer[counter].type_ = four_form; buffer[counter].alone_ = 1; 
				buffer[counter].form_ = medial;
				buffer[counter].vowel[0] = 0x00; strcpy(buffer[counter].cons,"I/"); counter++; }

<ARABIC>"||"(("a"|"i"|"u"|"a|")("N")?)?	{ purge(0); ECHO; }

<ARABIC>(("a"|"i"|"u"|"a|")("N")?)	{ if(counter==0) strcpy(buffer[counter-1].vowel,yytext); }

<ARABIC>"<>"	{ if (counter==0) strcpy(buffer[counter-1].vowel,"="); }

<ARABIC>"\\begin{"[A-Za-z]*"}"	{ purge(0); ECHO; }

<ARABIC>"\\end{"[A-Za-z]*"}"	{ purge(0); ECHO; }

<ARABIC>"\\ref{"[^}\n]*"}"	{ purge(0); ECHO; }

<ARABIC>"\\pageref{"[^}\n]*"}"	{ purge(0); ECHO; }

<ARABIC>"\\label{"[^}\n]*"}"	{ purge(0); ECHO; }

<ARABIC>"\\cite{"[^}\n]*"}"	{ purge(0); ECHO; }

<ARABIC>"\\cite["[^]\n]*"]{"[^}\n]*"}"	{ purge(0); ECHO; }

<ARABIC>\\[A-Za-z]*	{ purge(0); ECHO; }

<ARABIC>"$"("$")?	{ purge(0); ECHO; BEGIN IGNORE; }

<IGNORE>"%".*	ECHO;

<IGNORE>[^$]	ECHO;

<IGNORE>"$"("$")?	{ ECHO; BEGIN ARABIC; }

<ARABIC>([-+.])?[0-9]+([-+/.,])?[0-9]+	{ purge(0); printf("\\beginL"); ECHO; printf("\\endL{}"); }

<ARABIC>.|\n			{ purge(1); }
							
%%

void purge(int echoing)
{
	int i=0;
	
	perform_shaddas();
	perform_CA();
	if (urdu==1) urduize();
	perform_lam_alifs();
	if ((keshideh == 1) && (display==1)) { strcpy(keshideh_string,"\\kesh{}"); add_keshidehs(); } else { strcpy(keshideh_string,""); }
	if (debug != 0) { for (i=0;i<counter;i++)	{ printf("%%%%DEBUG> buffer[%d]: consonant '%s', vowel '%s', form '%d', keshideh %i\n",
		i,buffer[i].cons,buffer[i].vowel,buffer[i].form_,buffer[i].kesh_); }}
	print_buffer();
	initialize_buffers();
	if (echoing==1) ECHO;
}

void enter_arabic_mode(int is_display)
{
	display=is_display; 
	push_state(ARABIC); 
}

void print_buffer(void)
{
	int i=0;
	
	for (i=0;i<counter;i++) {
		if (((buffer[i].type_==two_form) && (buffer[i].form_==final)) || ((buffer[i].type_==four_form) && (buffer[i].form_==medial)))
		{ (buffer[i].kesh_==1) ? printf("%s%s%s",buffer[i].cons,buffer[i].vowel,keshideh_string) :
			printf("%s%s",buffer[i].cons,buffer[i].vowel); }
		else { (buffer[i].kesh_==1) ? printf("%s%i%s%s",buffer[i].cons,buffer[i].form_,buffer[i].vowel,keshideh_string) :
			printf("%s%i%s",buffer[i].cons,buffer[i].form_,buffer[i].vowel); }
		}
}

void add_keshidehs(void)
{
	int i=0;
	
	if (counter > 2) {
	for (i=0;i<counter-1;i++)	{ 
		if (((buffer[i].form_==2) || (buffer[i].form_==1)) && ((buffer[i+1].form_==2) || (buffer[i+1].form_==3)))
			buffer[i].kesh_=1;
		}
	}
}

void urduize(void)
{
	int i=0;
	
	for (i=0;i<counter;i++)	{ 
		if (!strcmp(buffer[i].cons,"-h")) strcpy(buffer[i].cons,"x");
		if (!strcmp(buffer[i].cons,"h")) { if (buffer[i].form_==final) strcat(buffer[i].cons,"/"); }
		}
}	

void perform_lam_alifs(void)
{
	int i=0,j=0;

	if (counter > 1) {
		for (i=0;i<counter-1;i++)	{ 
			if (!strcmp(buffer[i].cons,"l") && 
			(!strcmp(buffer[i+1].cons,"A") ||
			!strcmp(buffer[i+1].cons,"'a") ||
			!strcmp(buffer[i+1].cons,"'i") ||
			!strcmp(buffer[i+1].cons,"'A") ||
			!strcmp(buffer[i+1].cons,"AA")))
			{ 	strcat(buffer[i].cons,buffer[i+1].cons);
				buffer[i].form_= (buffer[i].form_ == initial) ? isolated : final; 
				buffer[i].type_=two_form; 
				if (buffer[i].vowel[0] != 0x00)	{ strcat(buffer[i].vowel,"/"); }
				strcat(buffer[i].vowel,buffer[i+1].vowel); 
				for (j=i+1;j<counter-1;j++) {
					strcpy(buffer[j].vowel,buffer[j+1].vowel);
					strcpy(buffer[j].cons,buffer[j+1].cons);
					buffer[j].form_=buffer[j+1].form_;
					buffer[j].type_=buffer[j+1].type_;
					}
				counter--;
				}
			}
		}
}

void perform_shaddas(void)
{
	int i=0,j=0;

	if (counter > 1) {
		for (i=0;i<counter-1;i++)	{ if ((buffer[i].vowel[0]==0x00) && (buffer[i+1].alone_==0) && (!strcmp(buffer[i].cons,buffer[i+1].cons)))
			{ 	strcpy(buffer[i].vowel,"+"); 
				strcat(buffer[i].vowel,buffer[i+1].vowel);
				for (j=i+1;j<counter-1;j++) {
					strcpy(buffer[j].vowel,buffer[j+1].vowel);
					strcpy(buffer[j].cons,buffer[j+1].cons);
					buffer[j].form_=buffer[j+1].form_;
					buffer[j].type_=buffer[j+1].type_;
					}
				counter--;
				}
			}
		}
}
		
void perform_CA(void)
{
	int i=0;
	
	for (i=0;i<counter;i++)	{ 
		if (i==0) {
			if (buffer[i].type_==two_form)	buffer[i].form_=isolated; else
				{ if (buffer[i].type_==four_form)	{ buffer[i].form_= (counter == 1) ? isolated : initial; }}
				}
		if ((buffer[i].type_==two_form)	&&
				(counter > i+1)) { buffer[i+1].form_ = (buffer[i+1].type_ == two_form) ? isolated : initial; }
		if ((counter == i+1) &&
				(buffer[i].type_ == four_form)) { buffer[i].form_ = (buffer[i].form_ == medial) ? final : isolated; }
		}
}

void initialize_buffers()
{
	int i=0;
	
	for (i=0;i<LENGTH;i++) { buffer[i].type_ = nothingT; buffer[i].form_ = isolated; 
		buffer[i].kesh_ = 0; buffer[i].alone_ = 0; buffer[i].cons[0] = 0x00; buffer[i].vowel[0] = 0x00; }
	counter=0;
}

void leave_arabic_mode()
{
	purge(0); pop_state(ARABIC);
}

void push_state(int new_state)
{
   if (state_stack_level >= MAX_STATE_STACK_SIZE) {
      fprintf(stderr,"ERROR: The stack of states is full!! I quit...");
      exit(0);
      }
   state_stack[++state_stack_level] = new_state;
   BEGIN state_stack[state_stack_level];
   
   switch (new_state) {
   	case 0:	fprintf(stderr,"Tentative d'empiler le mode 0 : ERREUR !!!\n"); break;
	case ARABIC: if (display==1) printf("\\arabpar{"); else printf("\\arab{"); break;
	case LATIN: printf("\\latin{"); break;
	default: break;
	}
}

void pop_state (int old_state)
{
   if (state_stack_level == -1) {
      fprintf(stderr,"ERROR: The stack of states is empty!! I quit...");
      exit(0);
      }
   state_stack_level--;
   BEGIN state_stack[state_stack_level];

   switch (old_state) {
   	case 0:	fprintf(stderr,"Tentative de quitter le mode 0 : ERREUR !!!\n"); break;
	case ARABIC: printf("}"); break;
	case LATIN: printf("}"); break;
	default: break;
	}
}

int yywrap(void)
{
	return 1;
}

main(int argc, char **argv, char **envv)
{
	int i=0;

#ifdef THINK_C
    argc = ccommand(&argv);
#endif
#ifdef _WILFRIED_
	InitStandalone (&argc, &argv, &envv);
#endif
	if (argc>1) {
		for (i=1;i<argc;i++)
		{
			if (!strcmp(argv[i],"-h"))
			{
				fprintf(stderr,"\n This is ArabicTeX, version %s\n (c) Yannis Haralambous 1995",VERSION);
				fprintf(stderr,"\n Syntax: arabitex [options] < input-file > output-file");
				fprintf(stderr,"\n Options are: ");
				fprintf(stderr,"\n           -h ---> is the help option you just asked");
				fprintf(stderr,"\n           -arabic <char between \\x80 and \\xFF> ---> Arabic mode escape char");
				fprintf(stderr,"\n           -latin <char between \\x80 and \\xFF> ---> Latin insertion mode escape char");
				fprintf(stderr,"\n           -d ---> debugging option");
				fprintf(stderr,"\n Default values are:");				
				fprintf(stderr,"\n            ---> Arabic mode escape char");
				fprintf(stderr,"\n            ---> Latin insertion mode escape char");
				fprintf(stderr,"\n Input/output data use standard streams (stdin/stdout)\n");
				exit(0);
			}
			else if (!strcmp(argv[i],"-arabic"))
			{
				arabic_delimiter=argv[++i];
			}
			else if (!strcmp(argv[i],"-latin"))
			{
				latin_delimiter=argv[++i];
			}
			else if (!strcmp(argv[i],"-d"))
			{
				debug=1;
			}
			else
			{
				fprintf(stderr,"\n Unknown option '%s': try '-h'",argv[i]);
			}
		}
	}
	initialize_buffers();
	yylex();
}